/**
 * Copyright 2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ajax4jsf.framework.ajax.xmlfilter;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Map;

import org.ajax4jsf.framework.util.javascript.JSEncoder;
import org.ajax4jsf.framework.util.message.Messages;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;


/**
 * @author shura
 * SAX content handler for serialise events as JavaScript function.
 */
public class JSContentHandler extends Writer implements ContentHandler,LexicalHandler {

    protected Writer outputWriter ;

    private static final JSEncoder encoder = new JSEncoder();

    private static final char S_EOL[] = System.getProperty("line.separator")
            .toCharArray();

    private static final char S_DOCUMENT_START[] = "(".toCharArray();

    private static final char S_DOCUMENT_ENF[] = ");".toCharArray();

    private static final char S_DOCUMENT_3[] = "\"?>".toCharArray();

    private static final char S_OBJECT_START[] = "{".toCharArray();
    private static final char[] S_OBJECT_END = "}".toCharArray();

    private static final char S_ELEMENT_START[] = "new E(".toCharArray();

    private static final char[] S_ELEMENT_CLOSE = ")".toCharArray();

    private static final char S_ATTRIBUTES_END[] = "}".toCharArray();

    private static final char S_ATTRIBUTES_START[] = ",{".toCharArray();

    private static final char S_ELEMENT_END_START_TAG[] = ",[".toCharArray();

    private static final char S_ELEMENT_END[] = "])".toCharArray();

    private static final char S_TEXT_START[] = "new T('".toCharArray();

    private static final char S_TEXT_END[] = "')".toCharArray();

    private static final char S_CDATA_START[] = "new D('".toCharArray();

    private static final char S_CDATA_END[] = "')".toCharArray();

    private static final char S_COMMENT_START[] = "new C('".toCharArray();

    private static final char S_COMMENT_END[] = "')".toCharArray();

    private static final char S_PROCINSTR_START[] = "<?".toCharArray();

    private static final char S_PROCINSTR_END[] = "?>".toCharArray();

    private static final char C_LT = '<';

    private static final char C_GT = '>';

    private static final char C_SPACE = ' ';

    private static final char C_QUOTE = '\'';

    private static final char C_NSSEP = ':';

    private static final char C_COMMA = ',';

    private static final boolean DEBUG = false;

    /* ====================================================================== */
    private boolean hanging_element = false;

    /** True if we are processing the prolog. */
    private boolean processing_prolog = true;

    /** True if we are processing the DTD. */
    private boolean processing_dtd = false;

    /** True if we are processing the DTD. */
    private boolean processing_cdata = false;
    /** The <code>DocType</code> instance representing the document. */
    // protected DocType doctype = null;

    protected int level;

    protected char[] indentBuffer;


    /**
     * <code>jsFunction</code>- name of JavaScript function to call
     */
    protected String jsFunction;

    /**
     * <code>parametersNames</code>- names of parameters for function
     */
    protected ArrayList parameters;

    private int currentParameter;

    /**
     * <code>src</code>- name of parameter for render as JavaScript object
     */
    private int src;

    private Locator locator;

    protected Map headers;

    protected String requestId;

    public static final String DEFAULT_ENCODING = "ISO-8859-1";



    
    public JSContentHandler() {
	}
	/**
     * @param outputWriter
     */
    public JSContentHandler(Writer outputWriter, String function,String requestId,Map headers) {
        this.outputWriter = outputWriter;
        this.jsFunction = function;
        this.headers = headers;
        this.requestId = requestId;
    }
    /**
     * @throws java.io.IOException
     */
    public void close() throws IOException {
        outputWriter.close();
    }
    /**
     * @throws java.io.IOException
     */
    public void flush() throws IOException {
        outputWriter.flush();
    }
    /**
     * @param cbuf
     * @param off
     * @param len
     * @throws java.io.IOException
     */
    public void write(char[] cbuf, int off, int len) throws IOException {
        outputWriter.write(cbuf, off, len);
    }
    
    // ContentHandler Methods
    
    
    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#characters(char[], int, int)
     */
    public void characters(char[] ch, int start, int length)
            throws SAXException {
        if (this.processing_prolog || level <= 0)
            return;
        try {
            if (!this.closeElement(false) && this.processing_cdata == false) {
                this.write(C_COMMA);
            }
            if (DEBUG) {
                this.write('[');
                this.write(Integer.toString(level));
                this.write(']');
            }
            if (this.processing_cdata == false) {
                this.write(S_TEXT_START);
            }
            this.encode(ch, start, length);
            if (this.processing_cdata == false) {
                this.write(S_TEXT_END);
            }
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }
    }
    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#endDocument()
     */
    public void endDocument() throws SAXException {
        if (this.level != 0) {
            throw new SAXException(Messages.getMessage(Messages.OPEN_CLOSE_TAGS_DO_NOT_MATCH_ERROR));
        }
        // Write parameters after parsing and final function )
        try {
            this.write(S_DOCUMENT_ENF);
            this.write(S_EOL);
            this.flush();
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }
    }
    
    
    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
     */
    public void endElement(String uri, String localName, String qName)
            throws SAXException {
        this.level--;
        if (closeElement(true))
            return;
        try {
            this.write(S_ELEMENT_END); // [</]
            if (DEBUG) {
                this.write('[');
                this.write(qName);
                this.write('-');
                this.write(uri);
                this.write(']');
            }
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }
        //        this.write(qual);
        //        this.write(C_GT); // [>]
    }

    /**
     * Write the end part of a start element (if necessary).
     * 
     * @param end_element
     *            Whether this method was called because an element is being
     *            closed or not.
     * @return <b>true </b> if this call successfully closed the element (and no
     *         further <code>&lt;/element&gt;</code> is required.
     */
    protected boolean closeElement(boolean end_element) throws SAXException {
        if (!hanging_element)
            return (false);
        try {
            if (end_element)
                this.write(S_ELEMENT_CLOSE); // [ />]
            else
                this.write(S_ELEMENT_END_START_TAG); // [>]
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }
        this.hanging_element = false;
        return (true);
    }

    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#endPrefixMapping(java.lang.String)
     */
    public void endPrefixMapping(String prefix) throws SAXException {
        // TODO Auto-generated method stub

    }
    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#ignorableWhitespace(char[], int, int)
     */
    public void ignorableWhitespace(char[] ch, int start, int length)
            throws SAXException {
        // TODO Auto-generated method stub

    }
    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#processingInstruction(java.lang.String, java.lang.String)
     */
    public void processingInstruction(String target, String data)
            throws SAXException {
        // TODO Auto-generated method stub

    }
    /* ====================================================================== */

    /**
     * Receive an object for locating the origin of SAX document events.
     */
    public final void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    /**
     * Return the public identifier for the current document event.
     *
     * @return A <code>String</code> containing the public identifier,
     *         or <b>null</b> if none is available.
     */
    public String getPublicId() {
        return(this.locator == null? null: this.locator.getPublicId());
    }

    /**
     * Return the system identifier for the current document event.
     *
     * @return A <code>String</code> containing the system identifier,
     *         or <b>null</b> if none is available.
     */
    public String getSystemId() {
        return(this.locator == null? null: this.locator.getSystemId());
    }
    
    /**
     * Return the line number where the current document event ends.
     *
     * @return The line number, or -1 if none is available.
     */
    public int getLineNumber() {
        return(this.locator == null? -1: this.locator.getLineNumber());
    }

    /**
     * Return the column number where the current document event ends.
     *
     * @return The column number, or -1 if none is available.
     */
    public int getColumnNumber() {
        return(this.locator == null? -1: this.locator.getColumnNumber());
    }

    /**
     * Return a <code>String</code> describing the current location.
     */
    protected String getLocation() {
        if (this.locator == null) return("");
        StringBuffer buf = new StringBuffer(" (");
        if (this.getSystemId() != null) {
            buf.append(this.getSystemId());
            buf.append(' ');
        }
        buf.append("line " + this.getLineNumber());
        buf.append(" col " + this.getColumnNumber());
        buf.append(')');
        return(buf.toString());
    }

    /* ====================================================================== */
    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#skippedEntity(java.lang.String)
     */
    public void skippedEntity(String name) throws SAXException {
        // TODO Auto-generated method stub

    }
    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#startDocument()
     */
    public void startDocument() throws SAXException {
        this.processing_prolog = false;
        this.processing_cdata = false;
        this.level = 0;
        // Write function name and arguments before processing ...
        try {
            this.write(jsFunction);
            this.write(S_DOCUMENT_START);
            this.write(S_EOL);
            // write request id
            this.write(C_QUOTE);
            this.write(requestId);
            this.write(C_QUOTE);
            this.write(C_COMMA);
            // write responce headers
            this.write(S_OBJECT_START);
            	for (Iterator keys = headers.entrySet().iterator(); keys.hasNext();) {
                    Map.Entry element = (Map.Entry) keys.next();
                    this.write(C_QUOTE);
                    encode( element.getKey().toString());
                    this.write(C_QUOTE);
                    this.write(C_NSSEP);
                    this.write(C_QUOTE);
                    encode( element.getValue().toString());
                    this.write(C_QUOTE);
                    if (keys.hasNext()) {
                        this.write(C_COMMA);
                    }
                }
            this.write(S_OBJECT_END);
            this.write(C_COMMA);
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }
        /* We have a document type. */
//        if (this.doctype != null) {
//
//            String root_name = this.doctype.getName();
//            /* Check the DTD and the root element */
//            if (!root_name.equals(qual)) {
//                throw new SAXException("Root element name \"" + root_name
//                        + "\" declared by document type declaration differs "
//                        + "from actual root element name \"" + qual + "\"");
//            }
            /* Output the <!DOCTYPE ...> declaration. */
            // this.write(this.doctype.toString());
//        }

        try {
            this.write(S_EOL);
        } catch (IOException e1) {
            throw new SAXException("Write error",e1);
        }
    }
    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
     */
    public void startElement(String uri, String localName,
            String qName, Attributes attributes) throws SAXException {
        try {
            if (!this.closeElement(false) && this.level > 0) {
                this.write(C_COMMA);
            }
            this.write(S_ELEMENT_START); // [<]
            if (DEBUG) {
                this.write('[');
                this.write(uri);
                this.write(']');
            }
            this.write(C_QUOTE);
            this.write(qName);
            this.write(C_QUOTE);

//            if (attributes.getLength() > 0) {
                this.write(S_ATTRIBUTES_START);
                // TODO - implementing namespaces !
                //        for (int x = 0; x < namespaces.length; x++) {
                //            this.write(S_ELEMENT_4); // [ xmlns]
                //            if (namespaces[x][Namespaces.NAMESPACE_PREFIX].length() > 0) {
                //                this.write(C_NSSEP); // [:]
                //                this.write(namespaces[x][Namespaces.NAMESPACE_PREFIX]);
                //            }
                //            this.write(S_ELEMENT_1); // [="]
                //            this.encode(namespaces[x][Namespaces.NAMESPACE_URI]);
                //            this.write(C_QUOTE); // ["]
                //        }

                for (int x = 0; x < attributes.getLength(); x++) {
                    this.write(C_SPACE); // [ ]
                    if (0 != x) {
                        this.write(C_COMMA);
                    }
                    if (DEBUG) {
                        this.write('[');
                        this.write(attributes.getURI(x));
                        this.write(']');
                    }
                    String attrName = attributes.getQName(x);
                    // For JavaScript any attributes names illegal ...
                    // replate with correct names.
                    if (attrName.equalsIgnoreCase("class")) {
                        attrName = "className";
                    }
                    this.write(attrName);
                    this.write(C_NSSEP); // [:]
                    this.write(C_QUOTE); // [']
                    this.encode(attributes.getValue(x));
                    this.write(C_QUOTE); // [']
                }

                this.write(S_ATTRIBUTES_END);
//            }
            this.level++;
            this.hanging_element = true;
        } catch (IOException e) {

            throw new SAXException("write error",e);
        }
    }
    /* (non-Javadoc)
     * @see org.xml.sax.ContentHandler#startPrefixMapping(java.lang.String, java.lang.String)
     */
    public void startPrefixMapping(String prefix, String uri)
            throws SAXException {
        // TODO Auto-generated method stub

    }
    
    /* (non-Javadoc)
     * @see org.xml.sax.ext.LexicalHandler#comment(char[], int, int)
     */
    public void comment(char[] ch, int start, int length) throws SAXException {
        if (this.processing_prolog || level <= 0)
            return;
        try {
            if (!this.closeElement(false)) {
                this.write(C_COMMA);
            }
            if (DEBUG) {
                this.write('[');
                this.write(Integer.toString(level));
                this.write(']');
            }
//            this.write(C_QUOTE);
            this.write(S_COMMENT_START);
            this.encode(ch, start, length);
            this.write(S_COMMENT_END);
//            this.write(C_QUOTE);
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }


    }
    /* (non-Javadoc)
     * @see org.xml.sax.ext.LexicalHandler#endCDATA()
     */
    public void endCDATA() throws SAXException {
        if (this.processing_prolog || level <= 0)
            return;
        try {
            if (this.closeElement(false)) {
                return;
            }
            if (DEBUG) {
                this.write('[');
                this.write(Integer.toString(level));
                this.write(']');
            }
            this.write(S_CDATA_END);
            this.processing_cdata = false;
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }
    }
    /* (non-Javadoc)
     * @see org.xml.sax.ext.LexicalHandler#endDTD()
     */
    public void endDTD() throws SAXException {
        // TODO Auto-generated method stub

    }
    /* (non-Javadoc)
     * @see org.xml.sax.ext.LexicalHandler#endEntity(java.lang.String)
     */
    public void endEntity(String name) throws SAXException {
        // TODO Auto-generated method stub

    }
    /* (non-Javadoc)
     * @see org.xml.sax.ext.LexicalHandler#startCDATA()
     */
    public void startCDATA() throws SAXException {
        if (this.processing_prolog || level <= 0)
            return;
        try {
            if (!this.closeElement(false)) {
                this.write(C_COMMA);
            }
            if (DEBUG) {
                this.write('[');
                this.write(Integer.toString(level));
                this.write(']');
            }
            this.write(S_CDATA_START);
            this.processing_cdata = true;
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }
    }
    /* (non-Javadoc)
     * @see org.xml.sax.ext.LexicalHandler#startDTD(java.lang.String, java.lang.String, java.lang.String)
     */
    public void startDTD(String name, String publicId, String systemId)
            throws SAXException {
        // TODO Auto-generated method stub

    }
    /* (non-Javadoc)
     * @see org.xml.sax.ext.LexicalHandler#startEntity(java.lang.String)
     */
    public void startEntity(String name) throws SAXException {
        // TODO Auto-generated method stub

    }
    protected boolean writeParameter(int num) throws SAXException {
        String parameterValue = (String) parameters.get(num);
        if (num == src) {
            return false;
        }
        if (null != parameterValue) {
            try {
                if (DEBUG) {
                    this.write('[');
                    this.write(num);
                    this.write(']');
                }
                this.write(C_QUOTE);
                this.write(parameterValue);
                this.write(C_QUOTE);
            } catch (IOException e) {
                throw new SAXException("write error",e);
            }
        }
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.apache.cocoon.components.serializers.EncodingSerializer#writeIndent(int)
     */
    protected void writeIndent(int indent) throws SAXException {
        try {
            this.write("\n".toCharArray(), 0, 1);
            if (indent > 0) {
                this.write(assureIndentBuffer(indent), 0, indent);
            }
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }

    }

    private char[] assureIndentBuffer(int size) {
        if (indentBuffer == null || indentBuffer.length < size) {
            indentBuffer = new char[size];
            Arrays.fill(indentBuffer, ' ');
        }
        return indentBuffer;
    }
    
    /**
     * Encode and write a <code>String</code>
     */
    protected void encode(String data)
    throws SAXException {
        char array[] = data.toCharArray();
        this.encode(array, 0, array.length);
    }
    
    /**
     * Encode and write an array of characters.
     */
    protected void encode(char data[])
    throws SAXException {
        this.encode(data, 0, data.length);
    }
    
    /**
     * Encode and write a specific part of an array of characters.
     */
    protected void encode(char data[], int start, int length)
    throws SAXException {
        int end = start + length;
        
        if (data == null) throw new NullPointerException("Null data");
        if ((start < 0) || (start > data.length) || (length < 0) ||
            (end > data.length) || (end < 0))
            throw new IndexOutOfBoundsException("Invalid data");
        if (length == 0) return;
        
        try {
            for (int x = start; x < end; x++) {
                char c = data[x];
                
                if (  JSContentHandler.encoder.compile(c)) {
                    continue;
                }
                
                if (start != x) this.write(data, start, x - start );
                this.write(JSContentHandler.encoder.encode(c));
                start = x + 1;
                continue;
            }
            if (start != end) this.write(data, start, end - start );
        } catch (IOException e) {
            throw new SAXException("Write error",e);
        }
    }
    
}
