/**
 * Licensed under the Common Development and Distribution License,
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.sun.com/cddl/
 *   
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package org.ajax4jsf.framework.resource;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.StreamCorruptedException;
import java.io.StringReader;
import java.net.URL;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import javax.faces.FacesException;
import javax.faces.context.FacesContext;
import javax.imageio.ImageIO;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;

import org.ajax4jsf.framework.util.base64.Codec;
import org.ajax4jsf.framework.util.config.WebXml;
import org.ajax4jsf.framework.util.message.Messages;
import org.apache.commons.digester.Digester;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Produce instances of InternetResource's for any types -
 * jar resource, dynamic created image, component-incapsulated etc.
 * Realised as singleton class to support cache, configuration etc.
 * 
 * @author shura (latest modification by $Author: alexsmirnov $)
 * @version $Revision: 1.6 $ $Date: 2007/01/08 17:37:29 $
 *
 */
public class ResourceBuilderImpl extends InternetResourceBuilder {
	
	private static final String DATA_SEPARATOR = "/DATA/";
	private static Map renderers;
	private static ResourceRenderer defaultRenderer = new MimeRenderer(null);
	
	/**
	 * keep resources instances .
	 * TODO - put this map to application-scope attribute, for support clastering environment.
	 */
	private Map resources = Collections.synchronizedMap(new HashMap());
	
	private long _startTime;
	
	private ServletContext servletContext;
	
	private WebXml webXml;
	private Codec codec;
	
	static {
		renderers = new HashMap();
		// append known renderers for extentions.
		renderers.put(".gif",new GifRenderer());
		ResourceRenderer renderer = new JpegRenderer();
		renderers.put(".jpeg",renderer);
		renderers.put(".jpg",renderer);
		renderers.put(".png",new PngRenderer());
		renderers.put(".js",new ScriptRenderer());
		renderers.put(".css",new StyleRenderer());
		renderers.put(".log",new LogfileRenderer());
		renderers.put(".html",new HTMLRenderer());
		renderers.put(".xhtml",new MimeRenderer("application/xhtml+xml"));
		renderers.put(".xml",new MimeRenderer("text/xml"));
		renderers.put(".xcss", new TemplateCSSRenderer());
//		renderers.put(".htc",new BehaviorRenderer());
		// set in-memory caching ImageIO
		ImageIO.setUseCache(false);

	}
	
	public WebXml getWebXml() {
		if(null == webXml){
			throw new FacesException("Resources framework is not initialised, check web.xml for Filter configuration");
		}
		return webXml;
	}

	public void setWebXml(WebXml webXml) {
		this.webXml = webXml;
	}
	
	public ResourceBuilderImpl() {
		super();
		_startTime = System.currentTimeMillis();
		InternetResourceBuilder.setInstance(this);
		registerResources();
	}

	  private void registerConfig(
			    URL resourceConfig)
			  {
			    try
			    {
			      InputStream in = resourceConfig.openStream();
			      try
			      {
			        Digester digester = new Digester();
			        digester.setValidating(false);
			        digester.setEntityResolver(new EntityResolver(){
				    	// Dummi resolver - alvays do nothing 
						public InputSource resolveEntity(String publicId, String systemId) throws SAXException, IOException {
							return new InputSource(new StringReader(""));
						}
			        });
				    digester.setNamespaceAware(false);
				    digester.setUseContextClassLoader(true);
			        digester.push(this);
			        digester.addObjectCreate("resource-config/resource", "class", JarResource.class);
			        digester.addObjectCreate("resource-config/resource/renderer", "class", HTMLRenderer.class);
			        digester.addCallMethod("resource-config/resource/renderer/content-type",
                            "setContentType", 0);
			        digester.addSetNext("resource-config/resource/renderer", "setRenderer", ResourceRenderer.class.getName());
			        digester.addCallMethod("resource-config/resource/name",
			                               "setKey", 0);
			        digester.addCallMethod("resource-config/resource/path",
			                               "setPath", 0);
			        digester.addCallMethod("resource-config/resource/cacheable",
			                               "setCacheable", 0);
			        digester.addCallMethod("resource-config/resource/session-aware",
                            				"setSessionAware", 0);
			        digester.addCallMethod("resource-config/resource/property",
			                               "setProperty", 2);
			        digester.addCallParam("resource-config/resource/property/name", 0);
			        digester.addCallParam("resource-config/resource/property/value", 1);
			        digester.addCallMethod("resource-config/resource/content-type",
			                               "setContentType", 0);
			        digester.addSetNext("resource-config/resource", "addResource", InternetResource.class.getName());
			        digester.parse(in);
			      }
			      finally
			      {
			        in.close();
			      }
			    }
			    catch (IOException e)
			    {
			      throw new FacesException(e);
			    }
			    catch (SAXException e)
			    {
			      throw new FacesException(e);
			    }
			  }
	/**
	 * @param servletContext
	 */
	public void init(ServletContext servletContext, String filterName) throws ServletException {
		this.servletContext = servletContext;
		setWebXml(new WebXml(servletContext,filterName));
		if ("true".equals(servletContext
				.getInitParameter(InternetResource.ENCODE_URI_PARAMETER))) {
			String random = servletContext
					.getInitParameter(InternetResource.ENCODE_PASS_PARAMETER);
			if (null == random) {
				random = "";
				for (int i = 0; i < 25; i++) {
					random += (char) (65 + 26 * Math.random());
				}
			}
			try {
				codec = new Codec(random);
			} catch (Exception e) {
				throw new ServletException("Error initialisation codec for resource data",e);
			}
		} else {
			codec = new Codec();
		}
		_startTime = System.currentTimeMillis();
	}

	/**
	 * @throws FacesException
	 */
	protected void registerResources() throws FacesException {
		try
	    {
	      ClassLoader loader = Thread.currentThread().getContextClassLoader();
	      Enumeration e = loader.getResources("META-INF/resources-config.xml");
	      while (e.hasMoreElements())
	      {
	        URL resource = (URL)e.nextElement();
	        registerConfig(resource);
	      }
	    }
	    catch (IOException e)
	    {
	      throw new FacesException(e);
	    }
	}

	/**
	 * Base point for creating resource. Must detect type and build
	 * appropriate instance. Currently - make static resource for ordinary
	 * request, or instance of class.
	 * @param base base object for resource ( resourcess in classpath will be get relative to it package )
	 * @param path key - path to resource, resource class name etc.
	 * @return 
	 * @throws FacesException 
	 */
	public InternetResource createResource(Object base, String path) throws  FacesException {
		// TODO - detect type of resource ( for example, resources location path in Skin
		try {
			return getResource(path);
		} catch (ResourceNotFoundException e) {
			try {
				return getResource(buildKey(base,path));
			} catch (ResourceNotFoundException e1) {
				if(log.isDebugEnabled()){
					log.debug(Messages.getMessage(Messages.BUILD_RESOURCE_INFO, path));
				}
			}
		}
		// path - is class name ?
		InternetResource res;
		try {
			Class resourceClass = Class.forName(path);
			res = createDynamicResource(path,resourceClass);
		} catch (Exception e) {
			try {
				res = createJarResource(base, path);
			} catch (ResourceNotFoundException ex) {
				res = createStaticResource(path);
			}
			// TODO - if resource not found, create static ?
		}		
		return res;
	}
	
	
	private String buildKey(Object base, String path) {
		if( path.startsWith("/")) {
			return path.substring(1);
		}
		if(null==base ){
			return path;
		}
		StringBuffer packageName = new StringBuffer(base.getClass().getPackage().getName().replace('.','/'));
		return packageName.append("/").append(path).toString();
	}
	
	public String getUri(InternetResource resource, FacesContext context, Object storeData){
		StringBuffer uri = new StringBuffer();// ResourceServlet.DEFAULT_SERVLET_PATH).append("/");
		uri.append(resource.getKey());
		// append serialized data as Base-64 encoded request string.
		if (storeData != null) {
			try {
				byte[] objectData;
				if( !(storeData instanceof byte[])){
				ByteArrayOutputStream dataSteram = new ByteArrayOutputStream(
						1024);
				ObjectOutputStream objStream = new ObjectOutputStream(
						dataSteram);
				objStream.writeObject(storeData);
				objStream.flush();
				objStream.close();
				dataSteram.close();
				objectData = dataSteram.toByteArray();
				} else {
					objectData = (byte[]) storeData;
				}
				uri.append(DATA_SEPARATOR);
				byte[] dataArray = encrypt(objectData);
				uri.append(new String(dataArray, "ISO-8859-1"));

				// / byte[] objectData = dataSteram.toByteArray();
				// / uri.append("?").append(new
				// String(Base64.encodeBase64(objectData),
				// / "ISO-8859-1"));
			} catch (Exception e) {
				// Ignore errors, log it
				log.error(Messages
						.getMessage(Messages.QUERY_STRING_BUILDING_ERROR), e);
			}
		}
		String resourceURL = getWebXml().getFacesResourceURL(context,
				uri.toString());// context.getApplication().getViewHandler().getResourceURL(context,uri.toString());
		if (resource.isSessionAware()) {
			resourceURL = context.getExternalContext().encodeResourceURL(
					resourceURL);
		}
		if (log.isDebugEnabled()) {
			log.debug(Messages.getMessage(Messages.BUILD_RESOURCE_URI_INFO,
					resource.getKey(), resourceURL));
		}
		return resourceURL;// context.getExternalContext().encodeResourceURL(resourceURL);

	}

	/**
	 * @param key
	 * @return
	 */
	public InternetResource getResourceForKey(String key) throws ResourceNotFoundException  {
		
		int data = key.indexOf(DATA_SEPARATOR);
		if(data>=0){
			key = key.substring(0,data);
		}
		return getResource(key);
	}
	
	public Object getResourceDataForKey(String key){
		Object data = null;
		String dataString = null;
			int dataStart = key.indexOf(DATA_SEPARATOR);
			if(dataStart>=0){
				dataString = key.substring(dataStart+DATA_SEPARATOR.length());
			}
			if (log.isDebugEnabled()) {
				log.debug(Messages.getMessage(
						Messages.RESTORE_DATA_FROM_RESOURCE_URI_INFO, key,
						dataString));
			}
			if (dataString != null) {
				// dataString =
				// dataString.substring(dataStart+ResourceServlet.DATA_PARAMETER.length()+1);
				byte[] objectArray = null;
				try {
					byte[] dataArray = dataString.getBytes("ISO-8859-1");
					objectArray = decrypt(dataArray);

					ObjectInputStream in = new ObjectInputStream(
							new ByteArrayInputStream(objectArray));
					data = in.readObject();
				} catch (StreamCorruptedException e) {
					data = objectArray;
				} catch (IOException e) {
					log.error(Messages
							.getMessage(Messages.DESERIALIZE_DATA_INPUT_ERROR), e);
				} catch (ClassNotFoundException e) {
					log.error(Messages
							.getMessage(Messages.DATA_CLASS_NOT_FOUND_ERROR), e);
				} 
			}
			return data;
	}

	public InternetResource getResource(String path) throws ResourceNotFoundException {

			InternetResource internetResource = (InternetResource)resources.get(path);
		if (null == internetResource) {
			throw new ResourceNotFoundException("Resource not registered : "+path);
		} else {
			return internetResource;
		}
	}
	
	public void addResource(InternetResource resource){
		resources.put(resource.getKey(), resource);
	}
	
	public void addResource(String key, InternetResource resource){
		resources.put(key,resource);
		resource.setKey(key);
		// TODO - set renderer ?
	}

	
	
	public String getFacesResourceKey(HttpServletRequest request) {
		return getWebXml().getFacesResourceKey(request);
	}

	public String getFacesResourceURL(FacesContext context, String Url) {
		return getWebXml().getFacesResourceURL(context, Url);
	}

	/**
	 * Build resource for link to static context in webapp.
	 * @param path
	 * @return
	 * @throws FacesException 
	 */
	protected  InternetResource createStaticResource(String path) throws ResourceNotFoundException, FacesException {
		FacesContext context = FacesContext.getCurrentInstance();
		if(null != context){
			if (context.getExternalContext().getContext() instanceof ServletContext) {
				ServletContext servletContext = (ServletContext) context.getExternalContext().getContext();
				InputStream in = servletContext.getResourceAsStream(path);
				if(null != in){
					InternetResourceBase res = new StaticResource(path);
					setRenderer(res,path);
					res.setLastModified(new Date(getStartTime()));
					addResource(path,res);
					try {
						in.close();
					} catch (IOException e) {
					}
					return res;					
				}
			}
		}
		throw new ResourceNotFoundException(Messages.getMessage(Messages.STATIC_RESOURCE_NOT_FOUND_ERROR, path));
	}

	private void setRenderer(InternetResourceBase res, String path) throws FacesException {
		int lastPoint = path.lastIndexOf('.');
		if(lastPoint > 0){
			String ext = path.substring(lastPoint);
			ResourceRenderer resourceRenderer = (ResourceRenderer) renderers.get(ext);
			if(null != resourceRenderer){
				res.setRenderer(resourceRenderer);
			} else {
				if(log.isDebugEnabled()){
					log.debug(Messages.getMessage(Messages.NO_RESOURCE_REGISTERED_ERROR_2, path, renderers.keySet()));
				}
				
//				String mimeType = servletContext.getMimeType(path);
				res.setRenderer(defaultRenderer);
			}
		}
	}

	/**
	 * Create resurce to send from classpath relative to base class. 
	 * @param base
	 * @param path
	 * @return
	 * @throws FacesException 
	 */
	protected InternetResource createJarResource(Object base, String path) throws ResourceNotFoundException, FacesException {
		String key = buildKey(base,path);
		ClassLoader loader = Thread.currentThread().getContextClassLoader();
		if(null != loader.getResource(key)){
		JarResource res = new JarResource(key);
		setRenderer(res,path);
		res.setLastModified(new Date(getStartTime()));
		addResource(key,res);
		return res;
		} else {
			throw new ResourceNotFoundException(Messages.getMessage(Messages.NO_RESOURCE_EXISTS_ERROR, key));
		}
		
	}
	
	/**
	 * Create resource by instatiate given class. 
	 * @param path
	 * @param instatiate
	 * @return
	 */
	protected InternetResource createDynamicResource(String path, Class instatiate) throws ResourceNotFoundException {
		if(InternetResource.class.isAssignableFrom(instatiate) ){
			InternetResource resource;
			try {
				resource = (InternetResource) instatiate.newInstance();
				addResource(path,resource);
			} catch (Exception e) {
				String message = Messages.getMessage(Messages.INSTANTIATE_RESOURCE_ERROR, instatiate.toString());
				log.error(message, e);
				throw new ResourceNotFoundException(message, e);
			}
			return resource;
		}
		throw new FacesException(Messages.getMessage(Messages.INSTANTIATE_CLASS_ERROR));
	}

	/**
	 * Create resource by instatiate {@link UserResource} class with given properties ( or got from cache ). 
	 * @param cacheable
	 * @param session
	 * @param mime
	 * @return
	 * @throws FacesException
	 */
	public InternetResource createUserResource(boolean cacheable, boolean session, String mime) throws  FacesException {
		String path = getUserResourceKey(cacheable, session, mime);
		InternetResource userResource;
		try {
			userResource = getResource(path);
		} catch (ResourceNotFoundException e) {
			userResource = new UserResource(cacheable,session,mime);
			addResource(path,userResource);
		}
		return userResource;
	}

	/**
	 * Generate resource key for user-generated resource with given properties.
	 * @param cacheable
	 * @param session
	 * @param mime
	 * @return
	 */
	private String getUserResourceKey(boolean cacheable, boolean session, String mime) {
		StringBuffer pathBuffer = new StringBuffer(UserResource.class.getName());
		pathBuffer.append(cacheable?"/c":"/n");
		pathBuffer.append(session?"/s":"/n");
		if(null != mime){
			pathBuffer.append('/').append(mime.hashCode());
		}
		String path = pathBuffer.toString();
		return path;
	}
	/**
	 * @return Returns the startTime for application.
	 */
	public long getStartTime() {
		return _startTime;
	}

	protected byte[] encrypt(byte[] src) {
		try {
			Deflater compressor = new Deflater(Deflater.BEST_SPEED);
			byte[] compressed = new byte[src.length + 100];
			compressor.setInput(src);
			compressor.finish();
			int totalOut = compressor.deflate(compressed);
			byte[] zipsrc = new byte[totalOut];
			System.arraycopy(compressed, 0, zipsrc, 0, totalOut);
			return codec.encode(zipsrc);
		} catch (Exception e) {
			throw new FacesException("Error encode resource data",e);
		}
	}

	protected byte[] decrypt(byte[] src) {
		try {
			byte[] zipsrc = codec.decode(src);
			Inflater decompressor = new Inflater();
			byte[] uncompressed = new byte[zipsrc.length * 5];
			decompressor.setInput(zipsrc);
			int totalOut = decompressor.inflate(uncompressed);
			decompressor.end();
			byte[] out = new byte[totalOut];
			System.arraycopy(uncompressed, 0, out, 0, totalOut);
			return out;
		} catch (Exception e) {
			throw new FacesException("Error decode resource data",e);
		}
	}

	
	
}
