/**
 * Licensed under the Common Development and Distribution License,
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.sun.com/cddl/
 *   
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package org.ajax4jsf.framework.resource.cached;

import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

import javax.faces.context.FacesContext;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;

import org.ajax4jsf.framework.resource.InternetResource;
import org.ajax4jsf.framework.resource.ResourceBuilderImpl;
import org.ajax4jsf.framework.resource.ResourceNotFoundException;
import org.ajax4jsf.framework.util.message.Messages;


/**
 * @author shura
 *
 */
public class CachedResourceBuilder extends ResourceBuilderImpl {

	private static final int DEFAULT_CAPACITY = 10000;

	private ServletContext servletContext;
	
	private long counter = 0;
	
	private DualLRUMap cache ;
	
	/* (non-Javadoc)
	 * @see org.ajax4jsf.framework.resource.ResourceBuilderImpl#decrypt(byte[])
	 */
	protected byte[] decrypt(byte[] data) {
		// dummy - data not send via internet.
		return data;
	}

	/* (non-Javadoc)
	 * @see org.ajax4jsf.framework.resource.ResourceBuilderImpl#encrypt(byte[])
	 */
	protected byte[] encrypt(byte[] data) {
		// dummy - data not send via internet.
		return data;
	}

	/* (non-Javadoc)
	 * @see org.ajax4jsf.framework.resource.ResourceBuilderImpl#getResourceDataForKey(java.lang.String)
	 */
	public Object getResourceDataForKey(String key) {
		ResourceBean bean = (ResourceBean) cache.get(key);
		if(null == bean){
			throw new ResourceNotFoundException("Resource for key "+key+"not present in cache");
		}
		return bean.getData();
	}

	/* (non-Javadoc)
	 * @see org.ajax4jsf.framework.resource.ResourceBuilderImpl#getResourceForKey(java.lang.String)
	 */
	public InternetResource getResourceForKey(String key) throws ResourceNotFoundException {
		ResourceBean bean = (ResourceBean) cache.get(key);
		if(null == bean){
			throw new ResourceNotFoundException("Resource for key "+key+"not present in cache");
		}
		return super.getResourceForKey(bean.getKey());
	}

	/* (non-Javadoc)
	 * @see org.ajax4jsf.framework.resource.ResourceBuilderImpl#getUri(org.ajax4jsf.framework.resource.InternetResource, javax.faces.context.FacesContext, java.lang.Object)
	 */
	public String getUri(InternetResource resource, FacesContext facesContext, Object data) {
		ResourceBean bean;
		if(null == data){
			bean = new ResourceBean(resource.getKey());
		} else {
			if (data instanceof byte[]) {
				// Special case for simple bytes array data.
				bean = new ResourceBytesDataBean(resource.getKey(),(byte[])data);
			} else {
				bean = new ResourceDataBean(resource.getKey(),data);
			}
		}
		String key = (String) cache.getKey(bean);
		if(null == key){
			synchronized (this) {
				counter++;
				key = bean.hashCode()+"c"+counter;
			}
			cache.put(key, bean);
		} else {
			// Refresh LRU
			cache.get(key);
		}
		String resourceURL = getFacesResourceURL(facesContext, key);
		if (resource.isSessionAware()) {
			resourceURL = facesContext.getExternalContext().encodeResourceURL(
					resourceURL);
		}
		if (log.isDebugEnabled()) {
			log.debug(Messages.getMessage(Messages.BUILD_RESOURCE_URI_INFO,
					resource.getKey(), resourceURL));
		}
		return resourceURL;// context.getExternalContext().encodeResourceURL(resourceURL);
	}

	/* (non-Javadoc)
	 * @see org.ajax4jsf.framework.resource.ResourceBuilderImpl#init(javax.servlet.ServletContext, java.lang.String)
	 */
	public void init(ServletContext servletContext, String filterName) throws ServletException {
		// TODO Auto-generated method stub
		super.init(servletContext, filterName);
		// Create cache manager.
		Properties properties = getProperties("cache.properties");
		int capacity = DEFAULT_CAPACITY;
		String capacityString = properties.getProperty("cache.capacity");
		if(null != capacityString){
			try {
			capacity = Integer.parseInt(capacityString);
			} catch (NumberFormatException e) {
				// use default.
			}
		}
		cache = new DualLRUMap(capacity);
		this.servletContext = servletContext;
		counter = getStartTime()-1158760000000L;
	}
	/**
	 * Get properties file from classpath
	 * 
	 * @param name
	 * @return
	 */
	protected Properties getProperties(String name) {
		Properties properties = new Properties();
		InputStream props = CachedResourceBuilder.class.getResourceAsStream(name);
		if (null != props) {
			try {
				properties.load(props);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				log.warn(Messages.getMessage(Messages.READING_PROPERTIES_ERROR,
						name), e);
			} finally {
				try {
					props.close();
				} catch (IOException e) {
					// Can be ignored
				}
			}
		}
		return properties;

	}
}
